from aspen.socket_io import client, utils


class XhrPolling(client.Client):

    def __init__(self):
        client.Client.__init__(self)
        self.timeout = None # no heartbeats
        self.closeTimeout = 8000
        self.duration = 20000

    def _onConnect(self, req, res):
        body = ''
        if req.method == 'GET':
            Client.prototype._onConnect.apply(self, [req, res])
            def foo():
                self._write('')
            self._closeTimeout = setTimeout(foo, self.options.duration)
            self._payload()
                
        else:
            assert req.method == 'POST'

            def on_body(message):
                body += message
            req.addListener('data', on_body)

            def on_end():
                headers = {'Content-Type': 'text/plain'}
                if (req.headers.origin):
                    if (self._verifyOrigin(req.headers.origin)):
                        headers.set('Access-Control-Allow-Origin', '*')
                        if (req.headers.cookie):
                            headers.set( 'Access-Control-Allow-Credentials'
                                       , 'true'
                                        )
                    else:
                        raise Response(401)
                try:
                    # optimization: just strip first 5 characters here?
                    msg = qs.parse(body)
                    self._onMessage(msg.data)
                except:
                    pass
    
                raise Response(200, "ok", headers)

            req.addListener('end', on_end)

    def _onClose(self):
        if (self._closeTimeout):
            clearTimeout(self._closeTimeout)
        return Client.prototype._onClose.call(self)
    
    def _write(self, message):
        if (self._open):
            headers = { 'Content-Type': 'text/plain; charset=UTF-8'
                      , 'Content-Length': len(message)
                       }
            # https://developer.mozilla.org/En/HTTP_Access_Control
            if (self.request.headers.origin and self._verifyOrigin(self.request.headers.origin)):
                headers.set('Access-Control-Allow-Origin', self.request.headers.origin)
                if (self.request.headers.cookie):
                    headers.set('Access-Control-Allow-Credentials', 'true')

            self._onClose()
            raise Response(200, message, headers)



"""URL rewriter for Aspen.

This module rewrites URLs based on request.path, where "route" means to alter
request.path and set a bunch of extra keys in environ. The file
,aspen/etc/routes.conf defines the routing configuration, with each line
containing one or two whitespace-separated tokens. The first token is a regular
expression to match against the incoming PATH_INFO. The second token, if
present, is the new request.path. If there is no second token, then any
matching requests will be permanently redirected to request.path plus a
trailing slash. Processing stops with the first match.

TODO: Sugar:
    /:foo/          foo.html
    =>
    /<?P(foo)[/^]+
    /<?P(foo)[/^]+/ foo.html

Groups matched in the regular expression are added to a dictionary at
request.urls.path.

By default, any direct requests for a path mentioned as a second token in
routes.conf will be responded to with 404. Use the A flag to [A]llow access.

"""
import logging
import os
import re
from os.path import isfile, join

import aspen
from aspen import restarter
from aspen.configuration.exceptions import ConfFileError
from aspen.http import Response


class Mapping(object):

    def __init__(self, pattern, path, flags):
        """Takes a regular expression pattern, a path, and flags.
        """
        self.pattern = re.compile(pattern)
        self.path = path
        self.flags

    def match(self, path):
        """Given a path, return a dictionary or None.

        The dictionary returned has keys corresponding to the named and unnamed
        groups in the match. Unnamed groups are keyed to their index as
        strings.

        """
        out = None
        match = self.pattern.match(path)
        if match is not None:
            matches = True
            out = match.groupdict()
            i = 0
            for group in match.groups():
                out[str(i)] = group
                i += 1
        return out 


# Startup Hook
# ============

mappings = []
protected = []

def startup(website):
    """Given a Website, configure the rewrite module.
    """

    confpath = join(website.root, '.aspen', 'etc', 'rewrite.conf')
    if not isfile(confpath):
        logging.info("No rewrite.conf")
    else:
        restarter.add(confpath)
        i = 0
        for line in open(confpath):
            i += 1
            line = line.split('#')[0].strip()
            if not line:
                continue

            parts = line.split() 
            nparts = len(parts)
            try:
                assert nparts in [1,2,3]
                if nparts == 1:     # will redirect to trailing slash
                    pattern = parts[0]
                    path = flags = ''
                elif nparts == 2:   # will serve new path
                    pattern, path = parts
                    flags = ''
                else:
                    assert nparts == 3 # sanity check
                    pattern, path, flags = parts
                    assert flags.startswith('[')
                    assert flags.endswith(']')
                    flags = flags[1:-1]
            except AssertionError:
                raise ConfFileError("Malformed line.", "rewrite.conf", i)

            mapping = Mapping(pattern, path, flags)
            mappings.append(mapping)
            if 'A' not in flags:
                protected.append(path)


# Inbound Hook
# ============

def inbound(request):
    path = request.path
    if path in protected:                       # protect direct access
        raise Response(404)
    for mapping in mappings:                    # look for a match
        match = mapping.match(path)
        if match is not None:
            request.original_path = path
            if mapping.path == '':              # redirect to trailing slash
                raise Response(301, headers={'Location': path+'/'}) #TODO qs
            else:                               # serve with explicit path
                request.path = mapping.path
            groups = match.groups()
            for key in range(len(groups)):      # add [i] indices to url
                val = groups[key]
                if val is None:
                    val = ''
                request.url[key] = val
            named = match.groupdict.items()
            for key, val in named:              # add ['name'] indices to url
                if val is None:
                    val = ''
                request.url[key] = val
            break                               # quit matching after first hit



import socket
import sys
try:
    import eventlet
    import eventlet.wsgi
except ImportError:
    print >> sys.stderr, ("Please install eventlet in order to "
                          "run aspen with the eventlet engine.")
    raise 

class DevNull:
    def write(self, msg):
        pass

def start(website):
    sock = eventlet.listen(website.address, website.sockfam)
    if eventlet.version_info <= (0, 9, 15):
        # Work around https://bitbucket.org/which_linden/eventlet/issue/86/
        if sys.platform[:3] != "win":
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    eventlet.wsgi.server(sock, website, log=DevNull())

def stop():
    pass

def start_restarter(check_all):
    def loop():
        while True:
            check_all()
            eventlet.sleep(0.5)
    eventlet.spawn_n(loop)



var Client = require('../client')
  , util = require(process.binding('natives').util ? 'util' : 'sys')
  , qs = require('querystring');

var Multipart = module.exports = function(){
  Client.apply(this, arguments);
};

util.inherits(Multipart, Client);

Multipart.prototype._onConnect = function(req, res){
  var self = this, body = '', headers = {};
  // https://developer.mozilla.org/En/HTTP_Access_Control
  if (req.headers.origin && this._verifyOrigin(req.headers.origin)){
    headers['Access-Control-Allow-Origin'] = '*';
    headers['Access-Control-Allow-Credentials'] = 'true';
  }
  if (typeof req.headers['access-control-request-method'] !== 'undefined'){
    // CORS preflight message
    headers['Access-Control-Allow-Methods'] = req.headers['access-control-request-method'];
    res.writeHead(200, headers);
    res.write('ok');
    res.end();
    return;
  }
  switch (req.method){
    case 'GET':
      Client.prototype._onConnect.apply(this, [req, res]);
      headers['Content-Type'] = 'multipart/x-mixed-replace;boundary="socketio"';
      headers['Connection'] = 'keep-alive';
      this.request.connection.addListener('end', function(){ self._onClose(); });
      this.response.useChunkedEncodingByDefault = false;
      this.response.shouldKeepAlive = true;
      this.response.writeHead(200, headers);
      this.response.write("--socketio\n");
      if ('flush' in this.response) this.response.flush();
      this._payload();
      break;
      
    case 'POST':
      headers['Content-Type'] = 'text/plain';
      req.addListener('data', function(message){
        body += message.toString();
      });
      req.addListener('end', function(){
        try {
          var msg = qs.parse(body);
          self._onMessage(msg.data);
        } catch(e){}
        res.writeHead(200, headers);
        res.write('ok');
        res.end();
        body = '';
      });
      break;
  }
};
  
Multipart.prototype._write = function(message){
  if (this._open){
    this.response.write("Content-Type: text/plain" + (message.length === 1 && message.charCodeAt(0) === 6 ? "; charset=us-ascii" : "") + "\n\n");
    this.response.write(message + "\n");
    this.response.write("--socketio\n");
  }
};



class WebSocketDisconnect(object):
    pass


def web_socket_handler(request, inq, outq):
    while True:
        try:
            ws_request = inq.get(timeout=0.5)
        except QueueTimeout:
            pass
        else:
            log.info("handling websocket request")
            log.info(ws_request)
            try:
                simplates.handle(request)
            except Response, response:
                pass
            except:
                response = Response(500, traceback.format_exc())
            outq.put(response)

def get_handshake(request):
    """Given a Request, return a handshake string.
    """
    ws_location = "ws://localhost" + request.path
    if request.headers.has('Sec-WebSocket-Key1'):
        protocol = request.headers.one('Sec-WebSocket-Protocol')
        key1 = request.headers.one('Sec-WebSocket-Key1')
        key2 = request.headers.one('Sec-WebSocket-Key2')
        key3 = receive(8)
        num1 = int(''.join(c for c in key1 if c in '0123456789'))
        num2 = int(''.join(c for c in key2 if c in '0123456789'))
        assert num1 % key1.count(' ') == 0
        assert num2 % key2.count(' ') == 0
        final = pack('!II8s', num1 / key1.count(' '), num2 / key2.count(' '), key3)
        out = (
'''HTTP/1.1 101 Web Socket Protocol Handshake\r\n
Upgrade: WebSocket\r\n
Connection: Upgrade\r\n
Sec-WebSocket-Origin: %s\r\n
Sec-WebSocket-Location: %s\r\n
'''% (org, ws_location))
        if protocol is not None:
            out += "Sec-WebSocket-Protocol: %s\r\n" % (protocol,)
        out += "\r\n"
        out += hashlib.md5(final).digest()
    else:
        out = (
'''HTTP/1.1 101 Web Socket Protocol Handshake\r
Upgrade: WebSocket\r
Connection: Upgrade\r
WebSocket-Origin: %s\r
WebSocket-Location: %s\r
WebSocket-Protocol: diesel-generic\r
\r
''' % (org, ws_location))

    return out


def websockets(request):
    """Abortive attempt to get websockets working.
    """
    org = request.headers.one('Origin')
    inq = Queue()
    outq = Queue()

    def wrap(request, inq, outq):
        web_socket_handler(request, inq, outq)
        outq.put(WebSocketDisconnect())
    fork(wrap, request, inq, outq)

    while True:
        try:
            log.debug("trying websocket thing")
            typ, val = first(receive=1, waits=[outq.wait_id])
            log.debug(typ)
            log.debug(val)
            if typ == 'receive':
                assert val == '\x00'
                val = until('\xff')[:-1]
                if val == '':
                    inq.put(WebSocketDisconnect())
                else:
                    inq.put(request)
            else:
                try:
                    v = outq.get(waiting=False)
                except QueueEmpty:
                    pass
                else:
                    if type(v) is WebSocketDisconnect:
                        send('\x00\xff')
                        break
                    else:
                        send('\x00%s\xff' % response.to_http(request.version))

        except ConnectionClosed:
            inq.put(WebSocketDisconnect())
            raise ConnectionClosed("remote disconnected")




var net = require('net')
  , util = require(process.binding('natives').util ? 'util' : 'sys')
  , WebSocket = require('./websocket')
  , listeners = []
  , netserver;

var Flashsocket = module.exports = function(){
  WebSocket.apply(this, arguments);
};

util.inherits(Flashsocket, WebSocket);

Flashsocket.httpUpgrade = true;

Flashsocket.init = function(listener){
  listeners.push(listener);
  
  listener.server.on('close', function(){
    listeners.splice(listeners.indexOf(listener), 1);

    if (listeners.length === 0 && netserver){
      try {
        netserver.close();
      } catch(e){}
    }
  });
  
  if (listener.options.flashPolicyServer && netserver === undefined){
    netserver = net.createServer(function(socket){
      socket.addListener('error', function(err){
        if (socket && socket.end){
          socket.end();
          socket.destroy();
        }
      });

      if(socket && socket.readyState == 'open')
        socket.end(policy(listeners));
    });
    
    try {
      netserver.listen(843);
    } catch(e){
      if (e.errno == 13)
        listener.options.log('Your node instance does not have root privileges. ' 
                           + 'This means that the flash XML policy file will be '
                           + 'served inline instead of on port 843. This will slow '
                           + 'down initial connections slightly.');
      netserver = null;
    }
  }
  
  // Could not listen on port 843 so policy requests will be inline
  listener.server.addListener('connection', function(stream){
    var flashCheck = function (data) {
      // Only check the initial data
      stream.removeListener("data", flashCheck);
      if (data[0] === 60 && data.length == 23) {
        if (data == '<policy-file-request/>\0') {
          listener.options.log("Answering flash policy request inline");
          if (stream && stream.readyState == 'open'){
            var xml = policy([listener]);
            stream.write(xml);
            stream.end();
          }
        }
      }
    };
    
    stream.on('data', flashCheck);
  });
};

function policy(listeners) {
  var xml = '<?xml version="1.0"?>\n<!DOCTYPE cross-domain-policy SYSTEM'
          + ' "http://www.macromedia.com/xml/dtds/cross-domain-policy.dtd">\n<cross-domain-policy>\n';

  listeners.forEach(function(l){
    [].concat(l.options.origins).forEach(function(origin){
      var parts = origin.split(':');
      xml += '<allow-access-from domain="' + parts[0] + '" to-ports="'+ parts[1] +'"/>\n';
    });
  });

  xml += '</cross-domain-policy>\n';
  return xml;
};



try:
    import json
except ImportError:
    import simplejson as json


frame = '~m~'

def stringify(message):
    if isinstance(message, basestring):
        return message
    else:
        return '~j~' + json.dumps(message)

def encode(messages):
    out = ''
    message = ''
    if isinstance(messages, basestring):
        messages = [messages]
    for message in messages:
        if not message:
            message = ''
        message = stringify(message)
        out += "%s%d%s%s" % (frame, len(message), frame, message)
    return out 

def decode(data):
    messages = []
    while data:
        if (data.substr(0, 3) != frame):
            return messages
        data = data.substr(3)
        number = ''
        n = ''
        for i in range(len(data)):
            n = Number(data.substr(i, 1))
            if (data.substr(i, 1) == n):
                number += n
            else:
                data = data.substr(number.length + frame.length)
                number = Number(number)
                break
        messages.push(data.substr(0, number)); # here
        data = data.substr(number)
    return messages



import logging
import os
import sys
import time
import threading
from os.path import join, isfile


log = logging.getLogger('aspen.restarter')
extras = []
mtimes = {}


def add(filename):
    extras.append(filename)

def check_one(filename):
    """Given a filename, return None or exit.
    """

    # The file may have been removed from the filesystem.
    # ===================================================

    if not isfile(filename):
        if filename in mtimes:
            log.info("file deleted: %s" % filename)
            sys.exit(1) # trigger restart
        else:
            # We haven't seen the file before. It has probably been loaded 
            # from a zip (egg) archive.
            return


    # Or not, in which case, check the mod time.
    # ==========================================

    mtime = os.stat(filename).st_mtime
    if filename not in mtimes: # first time we've seen it
        mtimes[filename] = mtime
    if mtime > mtimes[filename]:
        log.info("file changed: %s" % filename)
        sys.exit(1) # trigger restart

def check_all():
    """See if any of our available modules have changed on the filesystem.
    """
    for name, module in sorted(sys.modules.items()):    # module files
        filepath = getattr(module, '__file__', None)
        if filepath is None:
            # We land here when a module is an attribute of another module
            # i.e., it exists twice in the sys.modules table, once as its
            # canonical representation, and again having been imported
            # within another module.
            continue
        filepath = filepath.endswith(".pyc") and filepath[:-1] or filepath
        check_one(filepath)

    for filepath in extras:                             # additional files
        check_one(filepath)


# Setup
# =====

def install(website):
    """Given a Website instance, start a loop over check_all.
    """
    # This is not ideal. See http://sync.in/aspen-reloading

    if website.dotaspen is not None:
        for root, dirs, files in os.walk(website.dotaspen):
            for filename in files:
                add(join(root, filename))

    website.engine.start_restarter(check_all)



from aspen.http.mapping import Mapping


class Headers(Mapping):
    """Represent the headers in an HTTP Request message.
    """

    def __init__(self, headers):
        """Takes headers as a string.
        """
        Mapping.__init__(self)
        hd = {}
        for line in headers.splitlines():
            k, v = line.strip().split(': ', 1)
            hd[k.lower()] = v
        self._dict.update(hd)

    def to_http(self):
        """Return the headers as a string, formatted for an HTTP message.
        """
        out = []
        for header, values in self._dict.iteritems():
            for value in values:
                out.append('%s: %s' % (header.title(), value))
        return '\r\n'.join(out)


